/*******************************************************************
 * Xatm -- OCF, CID, TrueType scalable font support for X 
 *  By:  ͵ <y5t0964@students.chiba-u.ac.jp>
 */

/* This code is derived from X-VFlib's vffuncs.c */

/*$Id: Xatm_funcs.c,v 1.30 1999/03/26 01:08:09 elca Exp elca $*/

#include "Xatm.h"

/* X εư -deferglyphs all ץѤ뤳 */
#undef PreloadAllGlyphs

/*#define dprintf(str) fprintf str*/
#define dprintf(str) /* */

/*#define DEBUG*/
#undef DEBUG

/*#define DUMP*/
#undef DUMP

#ifdef DUMP
#include "dump.c"
#endif /* DUMP */

/**********************************************************************
 * ؿ
 */
static void convertNothing(XatmFontPtr ptr, char *p);
static void convertBitOrder(XatmFontPtr ptr, char *p);
static void convertByteOrder(XatmFontPtr ptr, char *p);
static void convertBitByteOrder(XatmFontPtr ptr, char *p);
static void XatmGetBitmap(FontPtr font,
			  unsigned int c,
			  unsigned int hbits,
			  unsigned int wbytes,
			  char *bits);
static int XatmMakeProps (FontInfoPtr ptr, FontScalablePtr vals);
static int XatmFont_Init(FontPtr font,
			 char* dir,
			 char* file,
			 double matrix[4]);
static void XatmFont_Deinit(FontPtr font);
static int XatmSetupRowCol(FontPtr         font,
			   FontEntryPtr    entry,
			   FontScalablePtr vals);
static int xatm_get_glyphs (FontPtr font,
			    unsigned long count,
			    unsigned char *chars,
			    FontEncoding encoding,
			    unsigned long *pCount,
			    CharInfoPtr *glyphs);
static int xatm_get_metrics (FontPtr font,
			     unsigned long count,
			     unsigned char *chars,
			     FontEncoding encoding,
			     unsigned long *pCount,
			     xCharInfo **glyphs);
static void xatm_unload_font (FontPtr font);
static void xatm_unload_glyphs (FontPtr font);

/***************************************************************************/

/*
 * bit/byte order convert functions
 */

static void convertNothing(XatmFontPtr ptr, char *p)
{
}

static void convertBitOrder(XatmFontPtr ptr, char *p)
{
	BitOrderInvert(p, ptr->wbytes * ptr->hbits);
}

static void convertByteOrder(XatmFontPtr ptr, char *p)
{
	if (ptr->font->bit != ptr->font->byte) {
		switch (ptr->font->scan) {
		case 1:
			break;
		case 2:
			TwoByteSwap(p, ptr->wbytes * ptr->hbits);
		case 4:
			FourByteSwap(p, ptr->wbytes * ptr->hbits);
		}
	}
}

static void convertBitByteOrder(XatmFontPtr ptr, char *p)
{
	convertBitOrder(ptr, p);
	convertByteOrder(ptr, p);
}

/*--------------------------------------------------------------*/
/* ghostscript ӥåȥޥåפ
 *
 * gs ٤Τǡʤ٤ô򤫤ʤ褦̤̿򸺤餹
 */
static void XatmGetBitmap(FontPtr font,
			  unsigned int c,
			  unsigned int hbits,
			  unsigned int wbytes,
			  char *bits)
{
	XatmFontPtr ptr = (XatmFontPtr) font->fontPrivate;
	char gscmd[512];
	char *tmp_bits = bits;
	int len, totallen=0;
	double x_delta, y_delta;
	int i;
	GsPipe gspipe;

	/* θƽФ¸ */
	static unsigned int b_hbits = 0, b_wbytes = 0;
	static char b_fontname[MAXFONTFILENAMELEN] = "\0";
	static double b_matrix[4] = {0.0, 0.0, 0.0, 0.0};


	/* ޡ */
	alarm(0);

	gspipe = gsGetPipe();

	/* gs->xfs Υѥץ饤Хåե򥯥ꥢ */
	xatm_flush(gspipe.r);

	/* ghostscript Υޥɤ */
	/* "16 16 .setpagesize /Wadalab-mincho-0-8 findfont
	   [16 0 16 16 0 0] makefont setfont 0 1 moveto
	   <3021> show shopage\n" -> "" */

	/* gs ô򸺤餹 xfs<=>gs ֤̤̿򸺤餹 */
	/* bitmap size */
	if (b_hbits != hbits || b_wbytes != wbytes || gspipe.init ) {
		sprintf(gscmd, "%d %d .setpagesize\n", wbytes * 8, hbits);
		write(gspipe.w, gscmd, strlen(gscmd));

		b_hbits = hbits;
		b_wbytes = wbytes;

		/*printf("page change: %s\n", gscmd);*/
	}
	/* font name & metric */
	if ((strcmp(b_fontname, ptr->fontname) != 0)
		|| b_matrix[0] != ptr->matrix[0]
		|| b_matrix[1] != ptr->matrix[1]
		|| b_matrix[2] != ptr->matrix[2]
		|| b_matrix[3] != ptr->matrix[3]
		|| gspipe.init ) {

		/* Ѵˤӥåȥޥåפ̳˽ФƤޤ
		   򤫤Ƹ褦ˤ */
		x_delta = computeXdelta(ptr->matrix);
		y_delta = computeYdelta(ptr->matrix)*0.75; /* 0.75 = Ĵο */

		sprintf(gscmd, "/%s findfont [%g %g %g %g %g %g] makefont setfont\n",
			ptr->fontname,
			ptr->matrix[0], ptr->matrix[1], ptr->matrix[2], ptr->matrix[3],
			x_delta + ptr->matrix[0] * ptr->delta_x,
			y_delta + ptr->matrix[3] * ptr->delta_y);
		write(gspipe.w, gscmd, strlen(gscmd));

		strcpy(b_fontname, ptr->fontname);
		b_matrix[0] = ptr->matrix[0];
		b_matrix[1] = ptr->matrix[1];
		b_matrix[2] = ptr->matrix[2];
		b_matrix[3] = ptr->matrix[3];

		/*printf("font change: %s\n", gscmd);*/
	}

	/* glyph */
	switch(ptr->glyph_byte) {
	case 2:
		sprintf(gscmd, "0 %g moveto <%04x> show showpage\n",
			hbits*0.156, c); /* 0.156 = Ĵο */
		break;
	case 1:
		sprintf(gscmd, "0 %g moveto <%02x> show showpage\n",
			hbits*0.156, c); /* 0.156 = Ĵο */
		break;
	default:
		fprintf(stderr, "unknown glyph width %s\n", ptr->fontname);
		break;
	}
	write(gspipe.w, gscmd, strlen(gscmd));

	dprintf((stdout, __FUNCTION__ ":gscmd %s\n", gscmd));
	dprintf((stdout, "hbits %d, wbytes %d\n", hbits, wbytes));

	/*usleep(500);*/

	/* XXX: ʸˤĤ 1ms * 10000 = 10 ǥॢ (X ǤޤΤɤ) */
	for (i = 0; i < 10000; i++) {
		len = read(gspipe.r, tmp_bits, hbits*wbytes);		
		if(gsGetPid() == 0) {
			fprintf(stderr, "gs crash. discard glyph->%s:%x\n",
				ptr->fontname, c);
			break;
		}
		if (len > 0) {
			totallen += len;
			tmp_bits += len;
		}
		if(totallen >= hbits * wbytes) {
			break;
		}

		usleep(1);
	}
	dprintf((stdout, "total %d\n", totallen));

	/* ޡ */
	alarm(gsGetWaitSec());
}


/************************************************************************
 * Call back functions
 */

static int xatm_get_glyphs (FontPtr font,
			    unsigned long count,
			    unsigned char *chars,
			    FontEncoding encoding,
			    unsigned long *pCount,
			    CharInfoPtr *glyphs)
{

	XatmFontPtr ptr = (XatmFontPtr) font->fontPrivate;
	CharInfoPtr charInfo;
	CharInfoPtr *glyphsBase = glyphs;
	unsigned int c;
	int hbits = ptr->hbits;
	int wbytes = ptr->wbytes;

	dprintf((stdout, "+" __FUNCTION__));

	switch (encoding) {
	case Linear8Bit:
	case TwoD8Bit:
		break;
	case Linear16Bit:
	case TwoD16Bit:
		while (count--) {
			c = *chars++ << 8;
			c |= *chars++;

			dprintf((stdout, "0x%x ", c));

                        if (!BTree_Search(ptr->btree, c, (void **)&charInfo)) {
                                charInfo = CharInfoPool_Get(ptr->pool);
                                charInfo->metrics = font->info.maxbounds;
				XatmGetBitmap(font, c, hbits, wbytes, charInfo->bits);
				(*ptr->convert)(ptr, charInfo->bits);
                                BTree_Insert(ptr->btree, c, charInfo);
                                /*BTree_Print(ptr->btree);*/

                        }
			*glyphs++ = charInfo;
		}
		break;
	}
	*pCount = glyphs - glyphsBase;

	dprintf((stdout, "pCount: %d\n", *pCount));
	dprintf((stdout, "-" __FUNCTION__ "()\n\n"));

	return Successful;

}

static int xatm_get_metrics (FontPtr font,
			     unsigned long count,
			     unsigned char *chars,
			     FontEncoding encoding,
			     unsigned long *pCount,
			     xCharInfo **glyphs)
{
	xCharInfo **glyphsBase = glyphs;
#ifdef PreloadAllGlyphs
	XatmFontPtr ptr = (XatmFontPtr) font->fontPrivate;
	CharInfoPtr charInfo;
	unsigned int c;
	int hbits = ptr->hbits;
	int wbytes = ptr->wbytes;
#endif

	dprintf((stdout, "\n+" __FUNCTION__ "(0x%x, %d, 0x%x, %d, 0x%x, 0x%x)\n",
		 font, count, chars, encoding ,count, glyphs));

	switch (encoding) {
	case Linear8Bit:
	case TwoD8Bit:
		break;
	case Linear16Bit:
	case TwoD16Bit:
		while (count--) {
#ifdef PreloadAllGlyphs
			c = *chars++ << 8;
			c |= *chars++;
			if (!BTree_Search(ptr->btree, c, (void **)&charInfo)) {
				charInfo = CharInfoPool_Get(ptr->pool);
				charInfo->metrics = font->info.maxbounds;
				XatmGetBitmap(font, c, hbits, wbytes, charInfo->bits);
				(*ptr->convert)(ptr, charInfo->bits);
				BTree_Insert(ptr->btree, c, charInfo);
			}
			*glyphs++ = &charInfo->metrics;
#else
                        /*
			  c = *chars++ << 8; c |= *chars++;
			  dprintf((stderr, "0x%x ", c));
			  */
			*glyphs++ = &font->info.maxbounds;
#endif /* PreloadAllGlyphs */
		}
		break;
	}
	*pCount = glyphs - glyphsBase;

	dprintf((stdout, "pCount: %d\n", *pCount));
	dprintf((stdout, "-" __FUNCTION__ "()\n\n"));

	return Successful;
}

static void xatm_unload_font (FontPtr font)
{
	dprintf((stdout, "\n+" __FUNCTION__ "(%x)\n", font));
	XatmFont_Deinit(font);
	dprintf((stdout, "-" __FUNCTION__ "()\n\n"));
}

static void xatm_unload_glyphs (FontPtr font)
{
	dprintf((stdout, "\n+" __FUNCTION__ "(%x)\n", font));
	dprintf((stdout, "-" __FUNCTION__ "()\n\n"));
}

/***************************************************************************/
/* ץѥƥϢ
 */
typedef struct {char *name; int flag; } PropTable;
#define A_NAME            0
#define A_NAME_LAST       1
#define A_NUMBER          2
#define A_NUMBER_PIXEL    3
#define A_NUMBER_POINT    4

static PropTable propTable[] = {
	{"FOUNDRY",          A_NAME},
	{"FAMILY_NAME",      A_NAME},
	{"WEIGHT_NAME",      A_NAME},
	{"SLANT",            A_NAME},
	{"SETWIDTH_NAME",    A_NAME},
	{"ADD_STYLE_NAME",   A_NAME},
	{"PIXEL_SIZE",       A_NUMBER_PIXEL},
	{"POINT_SIZE",       A_NUMBER_POINT},
	{"RESOLUTION_X",     A_NUMBER},
	{"RESOLUTION_Y",     A_NUMBER},
	{"SPACING",          A_NAME},
	{"AVERAGE_WIDTH",    A_NUMBER},
	{"CHARSET_REGISTRY", A_NAME},
	{"CHARSET_ENCODING", A_NAME_LAST},
};

/* propTable + "FONT" + "COPYRIGHT" + "X_HEIGHT" + "QUAD_WIDTH" = 18 */
#define NPROPS (XtNumber(propTable) + 4)

static int XatmMakeProps (FontInfoPtr ptr, FontScalablePtr vals)
{
	FontPropPtr props;
	int i, n = 0;
	char *is_str;
	char *cur, *next;

	dprintf((stdout, "\n+" __FUNCTION__ "()\n" ));

	/* props */
	ptr->props = (FontPropPtr) xalloc(sizeof(FontPropRec) * NPROPS);
	if (ptr->props == NULL) {
		return -1;
	}
	memset(ptr->props, 0, sizeof(FontPropRec) * NPROPS);
	props = ptr->props;

	/* isStringProp */
	ptr->isStringProp = (char *) xalloc(sizeof(char) * NPROPS);
	if (ptr->isStringProp == NULL) {
		xfree(ptr->props);
		return -1;
	}
	memset(ptr->isStringProp, 0, sizeof(char) * NPROPS);
	is_str = ptr->isStringProp;

	props[n].name = MakeAtom("FONT", sizeof("FONT") - 1, TRUE);
	props[n].value = MakeAtom(vals->xlfdName, strlen(vals->xlfdName), TRUE);
	*is_str = TRUE;

	is_str++;
	n++;

	cur = strchr(vals->xlfdName, '-') + 1;

	for (i = 0; i < XtNumber(propTable); i++) {

		dprintf((stdout, "."));


		next = strchr(cur, '-');
		if(next == NULL && propTable[i].flag != A_NAME_LAST) {
			break;
		}

		props[n].name = MakeAtom(propTable[i].name, strlen(propTable[i].name), TRUE);

		switch(propTable[i].flag) {
		case A_NAME: /* ʸ */
			*is_str = TRUE;
			if(next >= cur && *cur != '*') {
				props[n].value = MakeAtom(cur, next - cur, TRUE);
				dprintf((stdout, "Atom %s %s\n", propTable[i].name, cur));
			} else {
				props[n].value = MakeAtom(NULL, 0, TRUE);
			}
			break;
		case A_NAME_LAST: /* ʸ */
			*is_str = TRUE;
			if(*cur != '*') {
				props[n].value = MakeAtom(cur, strlen(cur), TRUE);
				dprintf((stdout, "Atom %s %s\n", propTable[i].name, cur));
			} else {
				props[n].value = MakeAtom(NULL, 0, TRUE);
			}
			break;
		case A_NUMBER: /*  */
			if(next >= cur && *cur != '*') {
				props[n].value = atoi(cur);
				dprintf((stdout, "num %d\n", atoi(cur)));
			} else {
				props[n].value = 0;
			}
			break;
		case A_NUMBER_PIXEL:
			props[n].value = d_to_i(computeChHeight(vals->pixel_matrix));
			break;
		case A_NUMBER_POINT:
			props[n].value = d_to_i(computeChHeight(vals->point_matrix) * 10.0);
			break;
		default:
			printf("prop unknown\n");
			break;
		}

		cur = next + 1;
		is_str++;
		n++;
	}

	props[n].name = MakeAtom("COPYRIGHT", sizeof("COPYRIGHT") - 1, TRUE);
	props[n].value = MakeAtom("Unknown: See Manual", sizeof("Unknown: See Manual") - 1, TRUE);
	*is_str = TRUE;

	is_str++;
	n++;

	props[n].name = MakeAtom("X_HEIGHT", sizeof("X_HEIGHT") - 1, TRUE);
	props[n].value = -1;

	is_str++;
	n++;

	props[n].name = MakeAtom("QUAD_WIDTH", sizeof("QUAD_WIDTH") - 1, TRUE);
	props[n].value = vals->width;

	is_str++;
	n++;

	/* nprops */
	ptr->nprops = n;

	dprintf((stdout, "n %d\n", n));
	dprintf((stdout, "-" __FUNCTION__ "()\n\n" ));
	return 0;
}

/***************************************************************************/

/*
 * initialize XatmFont struct
 */

static int XatmFont_Init(FontPtr font,
			 char* dir,
			 char* file,
			 double matrix[4])
{
	char fontName[MAXFONTFILENAMELEN];
	XatmFontPtr ptr;
	int wbytes;

	dprintf((stdout, "\n+XatmFont_Init(0x%x)\n", font));

	xatm_get_fontname(file, fontName);

	dprintf((stdout, "fontname %s\n", fontName));

	ptr = (XatmFontPtr) xalloc(sizeof(XatmFont));
	if (!ptr) {
		return AllocError;
	}

	ptr->fontname = (char *)xalloc(strlen(fontName) +1 );
	if(ptr->fontname == NULL) {
		xfree(ptr);
		return AllocError;
	}
	strcpy(ptr->fontname, fontName);


	ptr->font = font;

	/* 8 ӥåñ̤ˤ */
	wbytes = (d_to_i(computeWidth(matrix)) + 7) / 8;
	wbytes += font->glyph - 1;
	wbytes -= wbytes % font->glyph;
	ptr->wbytes = wbytes;

	ptr->hbits = d_to_i(computeHeight(matrix));

	ptr->matrix[0] = matrix[0];
	ptr->matrix[1] = matrix[1];
	ptr->matrix[2] = matrix[2];
	ptr->matrix[3] = matrix[3];

	ptr->delta_x = ptr->delta_y = 0.0;

	if(font->info.firstRow == 0 &&
	   font->info.lastRow == 0) {
		ptr->glyph_byte = 1;
	} else {
		ptr->glyph_byte = 2;
	}


	ptr->pool = CharInfoPool_Alloc(ptr->wbytes * ptr->hbits);
	if (ptr->pool == NULL) {
		xfree(ptr);
		return AllocError;
	}

	ptr->btree = BTree_Alloc();
	if (ptr->btree == NULL) {
		CharInfoPool_Free(ptr->pool);
		xfree(ptr);
		return AllocError;
	}

	if (font->bit == MSBFirst) {
		if (font->byte == LSBFirst) {
			ptr->convert = convertNothing;
		} else {
			ptr->convert = convertByteOrder;
		}
	} else {
		if (font->byte == LSBFirst) {
			ptr->convert = convertBitOrder;
		} else {
			ptr->convert = convertBitByteOrder;
		}
	}

	font->fontPrivate = (pointer) ptr;

	dprintf((stdout, "-XatmFont_Init()\n\n"));
	return Successful;
}


/* եȤɬפʤʤä */
static void XatmFont_Deinit(FontPtr font)
{
	XatmFontPtr ptr = (XatmFontPtr) font->fontPrivate;

	dprintf((stdout, "\n+XatmFont_Deinit(0x%x)\n", font));


	/* β */
	CharInfoPool_Free(ptr->pool);
	BTree_Free(ptr->btree);
	xfree(ptr->fontname);
	xfree(ptr);
	xfree(font->info.props);
	xfree(font->info.isStringProp);
	xfree(font);

	dprintf((stdout, "-XatmFont_Deinit()\n\n"));
}

/*************************************************************************/
typedef struct _rowcol {
	char * encode;
	int firstRow;
	int firstCol;
	int lastRow;
	int lastCol;
	int defaultCh;
} RowCol;

static RowCol xatm_row_col[] = {
	{"jisx0208",            0x21, 0x21, 0x74, 0x7e, 0x2121}, /* JISX0208 */
	{"jisx0212",            0x22, 0x21, 0x6d, 0x7e, 0x2221}, /* JISX0212 */
	{"big5",                0xa1, 0x00, 0xfa, 0xfe, 0xa140}, /* BIG5 */
	{"gb2312,1980",         0x21, 0x21, 0x77, 0x7e, 0x2121}, /* GB2312 */
	{"gb2312.80&gb8565.88", 0x21, 0x21, 0x7e, 0x7e, 0x2121}, /* GB2312/GB8565 */
	{"jisc6226",            0x21, 0x21, 0x77, 0x7e, 0x2121}, /* JISC6226 */
	{"ksc5601.1987-0",      0x21, 0x21, 0x7d, 0x7e, 0x2121}, /* KSC5601 */
	{"ksc5601.1987-1",      0xa1, 0xa1, 0xfd, 0xfe, 0xa1a1}, /* KSC5601(EUC) */
	{"cns11643.1986-1",     0x21, 0x21, 0x7d, 0x4b, 0x2121}, /* CNS11643(1) */
	{"cns11643.1986-2",     0x21, 0x21, 0x72, 0x44, 0x2121}, /* CNS11643(2) */
	{"jisx0201",            0x00, 0x20, 0x00, 0xff, 0x0020}, /* JISX0201 */
	{"iso8859",             0x00, 0x20, 0x00, 0xff, 0x0020}, /* ISO8859 */
	{"iso10646",            0x00, 0x00, 0xff, 0xff, 0x0020}, /* ISO10646/Unicode */
	{"x.hiragana",          0x00, 0x20, 0x00, 0x7e, 0x0020}, /* Xatm's own encoding */
	{"x.onebyte",           0x00, 0x01, 0x00, 0xff, 0x0020}, /* Xatm's own encoding */
	{"x.twobyte",           0x01, 0x01, 0xff, 0xff, 0x2121}, /* Xatm's own encoding */
};

#define ENCODE(encoding) (strcasestr(entry->name.name, encoding) || \
			  strcasestr(vals->xlfdName, encoding))
/* 饯¸ߤϰϤꤹ */
static int XatmSetupRowCol(FontPtr         font,
			   FontEntryPtr    entry,
			   FontScalablePtr vals)
{
	int i;

	for(i = 0; i < XtNumber(xatm_row_col); i++) {
		if(ENCODE(xatm_row_col[i].encode)) {
			/*printf(__FUNCTION__ ": %s\n", xatm_row_col[i].encode);*/
			font->info.firstRow = xatm_row_col[i].firstRow;
			font->info.firstCol = xatm_row_col[i].firstCol;
			font->info.lastRow  = xatm_row_col[i].lastRow;
			font->info.lastCol  = xatm_row_col[i].lastCol;
			font->info.defaultCh = xatm_row_col[i].defaultCh;
			return Successful;
		}
	}
	return BadFontName;
}
#undef ENCODE

/* scaleable font loader entry point */
int XatmOpenScalable(FontPathElementPtr fpe,
		     FontPtr	        *pFont,
		     int                flags,
		     FontEntryPtr       entry,
		     char               *fileName,
		     FontScalablePtr    vals,
		     fsBitmapFormat     format,
		     fsBitmapFormatMask fmask,
		     FontPtr            non_cachable_font) /* We don't do licensing */
{
	char xlfdName[MAXFONTNAMELEN];
	char fontName[MAXFONTFILENAMELEN];
	FontPtr font;
	int w, h, ch_w, ch_h, ascent, descent;
	int ret;
	int image;

	dprintf((stdout, "\n+" __FUNCTION__ "(%x, %x, %x, %x, %s, %x, %x, %x, %x)\n",
	 fpe, pFont, flags, entry, fileName, vals, format, fmask, non_cachable_font));
#ifdef DUMP
/*	DumpFontScalable(vals);
	DumpFontEntry(entry);*/
#endif

	strncpy(xlfdName, vals->xlfdName, sizeof(xlfdName));
	FontParseXLFDName(xlfdName, vals, FONT_XLFD_REPLACE_VALUE);

	dprintf((stdout, "\n" __FUNCTION__ ":%s -> %s\n\n", vals->xlfdName, xlfdName));
#if 0
	/* Reject outrageously small font sizes to keep the math from
	   blowing up. */
	if ((hypot(vals->pixel_matrix[0], vals->pixel_matrix[2]) < 1.0) ||
	    (hypot(vals->pixel_matrix[1], vals->pixel_matrix[3]) < 1.0)) {
		return BadFontName;
	}
#endif /* 0 */

	switch (entry->type) {
	case FONT_ENTRY_SCALABLE:
		strncpy(fontName, entry->u.scalable.fileName, sizeof(fontName));
		break;
	case FONT_ENTRY_SCALE_ALIAS:
		return BadFontName;
	case FONT_ENTRY_BITMAP:
		strncpy(fontName, entry->u.bitmap.fileName, sizeof(fontName));
		break;
	case FONT_ENTRY_ALIAS:
		return BadFontName;
	case FONT_ENTRY_BC:
		return BadFontName;
	default:
		return BadFontName;
	}

	font = (FontPtr) xalloc(sizeof(FontRec));
	if (!font) {
		return AllocError;
	}

	xatm_caps_change(entry, vals);

	/* 饯礭 */
	w = d_to_i(computeWidth(vals->pixel_matrix));
	h = d_to_i(computeHeight(vals->pixel_matrix));
	ch_w = d_to_i(computeChWidth(vals->pixel_matrix));
	ch_h = d_to_i(computeChHeight(vals->pixel_matrix));

	ascent = (h * 7) / 8;
	descent = h - ascent;

	vals->pixel = ch_h;
	vals->width = ch_w;

	/* 饯¸ߤϰϤꤹ */
	ret = XatmSetupRowCol(font, entry, vals);
	if (ret != Successful) {
		xfree(font);
		return ret;
	}

	font->refcnt = 0;
	font->info.noOverlap = 1;
	font->info.terminalFont = 1;
	font->info.constantMetrics = 1;
	font->info.constantWidth = 1;
	font->info.inkInside = 1;
	font->info.inkMetrics = 1;
	font->info.allExist = 0;
	font->info.drawDirection = LeftToRight;
	font->info.cachable = 1;
	font->info.anamorphic = 0;
	font->info.maxOverlap = 0;
	font->info.pad = (short)0xf0f0; /* 0, 0xf0f0 ??? */
	font->info.maxbounds.leftSideBearing = 0;
	font->info.maxbounds.rightSideBearing = w;
	font->info.maxbounds.characterWidth = ch_w;
	font->info.maxbounds.ascent = ascent;
	font->info.maxbounds.descent = descent;
	font->info.maxbounds.attributes = 0x00;
	font->info.minbounds.leftSideBearing = 0;
	font->info.minbounds.rightSideBearing = w;
	font->info.minbounds.characterWidth = ch_w;
	font->info.minbounds.ascent = ascent;
	font->info.minbounds.descent = descent;
	font->info.minbounds.attributes = 0x00;
	/* 10/14 : 0.7142 < 0.7857
	   11/16 : 0.6875 < 0.75
	   17/24 : 0.7083 < 0.75 */
	font->info.ink_maxbounds.leftSideBearing = (int)(w * 0.725);
	font->info.ink_maxbounds.rightSideBearing = w;
	font->info.ink_maxbounds.characterWidth = ch_w;
	font->info.ink_maxbounds.ascent = ascent;
	font->info.ink_maxbounds.descent = descent;
	font->info.ink_maxbounds.attributes = 0x00;
	font->info.ink_minbounds.leftSideBearing = 0;
	font->info.ink_minbounds.rightSideBearing = 0;
	font->info.ink_minbounds.characterWidth = ch_w;
	/* -1/14, -1/16,  -1/24 */
	font->info.ink_minbounds.ascent = -1; /* ??? */
	/* -11/14, -13/16  -21/24 */
	font->info.ink_minbounds.descent = -(ascent - 1); /* ??? */
	font->info.ink_minbounds.attributes = 0x00;
	font->info.fontAscent = ascent;
	font->info.fontDescent = descent;
	if (XatmMakeProps(&font->info, vals) < 0) {
		font->info.nprops = 0;
		font->info.props = NULL;
		font->info.isStringProp = NULL;
	}

        /*
	  font->bit = MSBFirst;
	  font->byte = LSBFirst;
	  font->glyph = 4;
	  font->scan = 1;
	  */
	font->scan = 1;

	FontDefaultFormat(&(font->bit), &(font->byte),
			  &(font->glyph), &(font->scan));
	ret = CheckFSFormat(format, fmask,
			    &(font->bit), &(font->byte),
			    &(font->scan), &(font->glyph),
			    &image);
	if (ret != Successful) {
		xfree(font);
		return ret;
	}
	font->format = format;
	font->get_glyphs = xatm_get_glyphs;
	font->get_metrics = xatm_get_metrics;
	font->unload_font = xatm_unload_font;
	font->unload_glyphs = xatm_unload_glyphs;
        /*
	  font->fpe = ;
	  font->svrPrivate = ;
	  */
        /*
	  font->fontPrivate = ;
	  */


	ret = XatmFont_Init(font, fpe->name, entry->u.scalable.fileName, vals->pixel_matrix);
	if (ret != Successful) {
		xfree(font);
		return ret;
	}
	xatm_caps_change_xatm(entry, (XatmFontPtr)font->fontPrivate);

        /*
	  font->fpePrivate = ;
	  */
	font->maxPrivate = -1;
	font->devPrivates = NULL;


#ifdef DUMP
	/*DumpFont(font);*/
#endif

	*pFont = font;

	dprintf((stdout, "FontPtr 0x%x\n", *pFont));

	dprintf((stdout, "-XatmOpenScalable()\n\n"));
	return Successful;

}

int XatmGetInfoScalable(FontPathElementPtr  fpe,
			FontInfoPtr	    pFontInfo,
			FontEntryPtr        entry,
			FontNamePtr	    pFontName,
			char	            *fileName,
			FontScalablePtr	    vals)
{
	FontPtr pfont;
	int flags = 0;
	long format = 0;  /* It doesn't matter what format for just info */
	long fmask = 0;
	int ret;

	dprintf((stderr, "\nXatmGetInfoScalable(%x, %x, %x, %x, %s, %x)\n",
		 fpe, pFontInfo, entry, pFontName, fileName, vals));

	ret = XatmOpenScalable(fpe, &pfont, flags, entry, fileName, vals,
				   format, fmask, NULL);
	if (ret != Successful)
		return ret;
	*pFontInfo = pfont->info;

	(*pfont->unload_font)(pfont);

	return Successful;
}


static FontRendererRec renderer = {
	fileSuffix:      XATM_EXT,
	fileSuffixLen:   XATM_EXT_LEN,
	OpenBitmap:      NULL,
	OpenScalable:    XatmOpenScalable,
	GetInfoBitmap:   NULL,
	GetInfoScalable: XatmGetInfoScalable,
	number:          0,
	capabilities:    CAP_MATRIX | CAP_CHARSUBSETTING,
};

void XatmRegisterFontFileFunctions(void)
{
	struct sigaction s_act;

	dprintf((stdout, "\n+XatmRegisterFontFileFunctions()\n"));

	s_act.sa_flags = 0;
	sigemptyset(&(s_act.sa_mask));
	sigaddset(&(s_act.sa_mask), SIGALRM);
	sigaddset(&(s_act.sa_mask), SIGCHLD);

	s_act.sa_handler = gsSigHandler;
	sigaction(SIGCHLD, &s_act, NULL);
	sigaction(SIGALRM, &s_act, NULL);

	atexit(gsDeinit);

	FontFileRegisterRenderer(&renderer);

	dprintf((stdout, "-XatmRegisterFontFileFunctions()\n\n"));
}

/* end */
